import { createFilter } from 'rollup-pluginutils';
import { findConfigFile, sys, getParsedCommandLineOfConfigFile, createProgram, nodeModuleNameResolver, createSourceFile, ScriptTarget, ModifierFlags, isFunctionDeclaration, isInterfaceDeclaration, isClassDeclaration, isTypeAliasDeclaration, isVariableStatement, isExportDeclaration, isExportAssignment, isImportDeclaration, SyntaxKind, isIdentifier, isNamedImports, isLiteralExpression, getCombinedModifierFlags, isPropertyDeclaration, isPropertySignature, isIndexSignatureDeclaration, isMethodDeclaration, isMethodSignature, isConstructorDeclaration, isConstructSignatureDeclaration, isParenthesizedTypeNode, isTypeOperatorNode, isArrayTypeNode, isTupleTypeNode, isUnionTypeNode, isIntersectionTypeNode, isTypeLiteralNode, isMappedTypeNode, isConditionalTypeNode, isIndexedAccessTypeNode, isTypeReferenceNode } from 'typescript';
import path from 'path';

const MARKER = "0";
let IDs = 1;
/**
 * Create a new `Program` for the given `node`:
 */
function createProgram$1(node) {
    return withStartEnd({
        type: "Program",
        sourceType: "module",
        body: [],
    }, node);
}
/**
 * Create an export for `id`:
 * `export default id`
 */
function createDefaultExport(node, range) {
    const id = createIdentifier(node, range);
    return withStartEnd({
        type: "ExportDefaultDeclaration",
        declaration: id,
    }, range);
}
/**
 * Create an export for `id`:
 * `export { id }`
 */
function createExport(node, range) {
    const id = createIdentifier(node, range);
    return withStartEnd({
        type: "ExportNamedDeclaration",
        declaration: null,
        specifiers: [
            {
                type: "ExportSpecifier",
                exported: id,
                local: id,
            },
        ],
    }, range);
}
/**
 * Creates a reference to `id`:
 * `_ = ${id}`
 */
function createReference(id) {
    return {
        type: "AssignmentPattern",
        left: {
            type: "Identifier",
            name: String(IDs++),
        },
        right: id,
    };
}
function createIdentifier(node, range = node) {
    return withStartEnd({
        type: "Identifier",
        name: node.text || String(node.escapedText),
    }, range);
}
/**
 * Create a new Declaration and Scope for `id`:
 * `function ${id}(_ = MARKER) {}`
 */
function createDeclaration(id, range) {
    const start = getStart(range);
    return withStartEnd({
        type: "FunctionDeclaration",
        id: withStartEnd({
            type: "Identifier",
            name: id.text,
        }, { start, end: start }),
        params: [createReference(createIdentifier(id))],
        body: { type: "BlockStatement", body: [] },
    }, range);
}
function convertExpression(node) {
    if (isLiteralExpression(node)) {
        return { type: "Literal", value: node.text };
    }
    // istanbul ignore else
    if (isIdentifier(node)) {
        return createIdentifier(node);
    }
    else {
        console.log({ kind: node.kind, code: node.getFullText() });
        throw new Error(`Unknown Expression`);
    }
}
/**
 * Mark the nested `range` to be removed, by creating dead code:
 * `_ = () => {MARKER}`
 */
function removeNested(range) {
    return createReference({
        type: "FunctionExpression",
        id: null,
        body: withStartEnd({
            type: "BlockStatement",
            body: [
                withStartEnd({
                    type: "ExpressionStatement",
                    expression: { type: "Identifier", name: MARKER },
                }, range),
            ],
        }, {
            start: getStart(range) - 1,
            end: range.end + 1,
        }),
        params: [],
    });
}
function getStart({ start, pos }) {
    return typeof start === "number" ? start : pos || 0;
}
function withStartEnd(node, range = { start: 0, end: 0 }) {
    return Object.assign(node, {
        start: getStart(range),
        end: range.end,
    });
}
function matchesModifier(node, flags) {
    const nodeFlags = getCombinedModifierFlags(node);
    return (nodeFlags & flags) === flags;
}

const IGNORE_TYPENODES = new Set([
    SyntaxKind.LiteralType,
    SyntaxKind.VoidKeyword,
    SyntaxKind.UnknownKeyword,
    SyntaxKind.AnyKeyword,
    SyntaxKind.BooleanKeyword,
    SyntaxKind.NumberKeyword,
    SyntaxKind.StringKeyword,
    SyntaxKind.ObjectKeyword,
    SyntaxKind.NullKeyword,
    SyntaxKind.UndefinedKeyword,
    SyntaxKind.SymbolKeyword,
    SyntaxKind.NeverKeyword,
    SyntaxKind.ThisKeyword,
]);
class DeclarationScope {
    constructor(id, range) {
        this.declaration = createDeclaration(id, range);
    }
    pushRaw(expr) {
        this.declaration.params.push(expr);
    }
    pushReference(id) {
        this.pushRaw(createReference(id));
    }
    pushIdentifierReference(id) {
        this.pushReference(createIdentifier(id));
    }
    removeModifier(node, kind = SyntaxKind.ExportKeyword) {
        for (const mod of node.modifiers || []) {
            if (mod.kind === kind) {
                const start = node.getStart();
                const end = mod.end + 1;
                this.pushRaw(removeNested({ start, end }));
            }
        }
    }
    convertParametersAndType(node) {
        for (const param of node.parameters) {
            if (param.type) {
                this.convertTypeNode(param.type);
            }
        }
        if (node.type) {
            this.convertTypeNode(node.type);
        }
    }
    convertHeritageClauses(node) {
        for (const heritage of node.heritageClauses || []) {
            for (const type of heritage.types) {
                this.pushReference(convertExpression(type.expression));
            }
        }
    }
    convertMembers(members) {
        for (const node of members) {
            if ((isPropertyDeclaration(node) || isPropertySignature(node) || isIndexSignatureDeclaration(node)) &&
                node.type) {
                this.convertTypeNode(node.type);
            }
            // istanbul ignore else
            else if (isMethodDeclaration(node) ||
                isMethodSignature(node) ||
                isConstructorDeclaration(node) ||
                isConstructSignatureDeclaration(node)) {
                this.convertParametersAndType(node);
            }
            else {
                console.log({ kind: node.kind, code: node.getFullText() });
                throw new Error(`Unknown TypeElement`);
            }
        }
    }
    convertTypeNode(node) {
        if (IGNORE_TYPENODES.has(node.kind)) {
            return;
        }
        if (isParenthesizedTypeNode(node) || isTypeOperatorNode(node)) {
            return this.convertTypeNode(node.type);
        }
        if (isArrayTypeNode(node)) {
            return this.convertTypeNode(node.elementType);
        }
        if (isTupleTypeNode(node)) {
            for (const type of node.elementTypes) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isUnionTypeNode(node) || isIntersectionTypeNode(node)) {
            for (const type of node.types) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isTypeLiteralNode(node)) {
            return this.convertMembers(node.members);
        }
        if (isMappedTypeNode(node)) {
            const { typeParameter, type } = node;
            if (typeParameter.constraint) {
                this.convertTypeNode(typeParameter.constraint);
            }
            // TODO: create scopes for the name
            // node.typeParameter.name
            if (type) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isConditionalTypeNode(node)) {
            this.convertTypeNode(node.checkType);
            // TODO: create scopes for `infer`
            this.convertTypeNode(node.extendsType);
            this.convertTypeNode(node.trueType);
            this.convertTypeNode(node.falseType);
            return;
        }
        if (isIndexedAccessTypeNode(node)) {
            this.convertTypeNode(node.objectType);
            this.convertTypeNode(node.indexType);
            return;
        }
        // istanbul ignore else
        if (isTypeReferenceNode(node)) {
            return this.pushReference(this.convertEntityName(node.typeName));
        }
        else {
            console.log({ kind: node.kind, code: node.getFullText() });
            throw new Error(`Unknown TypeNode`);
        }
    }
    convertEntityName(node) {
        if (isIdentifier(node)) {
            return createIdentifier(node);
        }
        return withStartEnd({
            type: "MemberExpression",
            computed: false,
            object: this.convertEntityName(node.left),
            property: createIdentifier(node.right),
        }, 
        // TODO: clean up all the `start` handling!
        { start: node.getStart(), end: node.end });
    }
}

class Transformer {
    constructor(sourceFile) {
        this.ast = createProgram$1(sourceFile);
        for (const stmt of sourceFile.statements) {
            this.convertStatement(stmt);
        }
    }
    transform() {
        return this.ast;
    }
    pushStatement(node) {
        this.ast.body.push(node);
    }
    maybeMarkAsExported(node, id) {
        if (matchesModifier(node, ModifierFlags.ExportDefault)) {
            const start = node.pos;
            this.pushStatement(createDefaultExport(id, { start, end: start }));
        }
        else if (matchesModifier(node, ModifierFlags.Export)) {
            const start = node.pos;
            this.pushStatement(createExport(id, { start, end: start }));
        }
    }
    createDeclaration(id, range) {
        const scope = new DeclarationScope(id, range);
        this.pushStatement(scope.declaration);
        return scope;
    }
    convertStatement(node) {
        if (isFunctionDeclaration(node)) {
            return this.convertFunctionDeclaration(node);
        }
        if (isInterfaceDeclaration(node) || isClassDeclaration(node)) {
            return this.convertClassOrInterfaceDeclaration(node);
        }
        if (isTypeAliasDeclaration(node)) {
            return this.convertTypeAliasDeclaration(node);
        }
        if (isVariableStatement(node)) {
            return this.convertVariableStatement(node);
        }
        if (isExportDeclaration(node) || isExportAssignment(node)) {
            return this.convertExportDeclaration(node);
        }
        // istanbul ignore else
        if (isImportDeclaration(node)) {
            return this.convertImportDeclaration(node);
        }
        else {
            console.log({ kind: node.kind, code: node.getFullText() });
            throw new Error(`unsupported node type`);
        }
    }
    convertFunctionDeclaration(node) {
        // istanbul ignore if
        if (!node.name) {
            console.log({ code: node.getFullText() });
            throw new Error(`FunctionDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
        scope.convertParametersAndType(node);
    }
    convertClassOrInterfaceDeclaration(node) {
        // istanbul ignore if
        if (!node.name) {
            console.log({ code: node.getFullText() });
            throw new Error(`ClassDeclaration / InterfaceDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        if (isInterfaceDeclaration(node)) {
            scope.removeModifier(node, SyntaxKind.DefaultKeyword);
        }
        scope.convertHeritageClauses(node);
        // NOTE(swatinem): typescript loses the non-null assertion for `node.name`
        scope.pushIdentifierReference(node.name);
        scope.convertMembers(node.members);
    }
    convertTypeAliasDeclaration(node) {
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
        scope.convertTypeNode(node.type);
    }
    convertVariableStatement(node) {
        const { declarations } = node.declarationList;
        // istanbul ignore if
        if (declarations.length !== 1) {
            console.log({ code: node.getFullText() });
            throw new Error(`VariableStatement with more than one declaration not yet supported`);
        }
        for (const decl of declarations) {
            // istanbul ignore if
            if (!isIdentifier(decl.name)) {
                console.log({ code: node.getFullText() });
                throw new Error(`VariableDeclaration must have a name`);
            }
            this.maybeMarkAsExported(node, decl.name);
            const scope = this.createDeclaration(decl.name, node);
            scope.removeModifier(node);
            scope.pushIdentifierReference(decl.name);
            if (decl.type) {
                scope.convertTypeNode(decl.type);
            }
        }
    }
    convertExportDeclaration(node) {
        if (isExportAssignment(node)) {
            this.pushStatement(withStartEnd({
                type: "ExportDefaultDeclaration",
                declaration: convertExpression(node.expression),
            }, node));
            return;
        }
        const source = node.moduleSpecifier ? convertExpression(node.moduleSpecifier) : undefined;
        if (!node.exportClause) {
            this.pushStatement(withStartEnd({
                type: "ExportAllDeclaration",
                source,
            }, node));
        }
        else {
            this.pushStatement(withStartEnd({
                type: "ExportNamedDeclaration",
                declaration: null,
                specifiers: node.exportClause ? node.exportClause.elements.map(e => this.convertExportSpecifier(e)) : [],
                source,
            }, node));
        }
    }
    convertImportDeclaration(node) {
        const source = convertExpression(node.moduleSpecifier);
        // istanbul ignore if
        if (!node.importClause || (!node.importClause.name && !node.importClause.namedBindings)) {
            console.log({ code: node.getFullText() });
            throw new Error(`ImportDeclaration should have imports`);
        }
        const specifiers = node.importClause.namedBindings
            ? this.convertNamedImportBindings(node.importClause.namedBindings)
            : [];
        if (node.importClause.name) {
            specifiers.push({
                type: "ImportDefaultSpecifier",
                local: createIdentifier(node.importClause.name),
            });
        }
        this.pushStatement(withStartEnd({
            type: "ImportDeclaration",
            specifiers,
            source,
        }, node));
    }
    convertNamedImportBindings(node) {
        if (isNamedImports(node)) {
            return node.elements.map(el => {
                const local = createIdentifier(el.name);
                const imported = el.propertyName ? createIdentifier(el.propertyName) : local;
                return {
                    type: "ImportSpecifier",
                    local,
                    imported,
                };
            });
        }
        return [
            {
                type: "ImportNamespaceSpecifier",
                local: createIdentifier(node.name),
            },
        ];
    }
    convertExportSpecifier(node) {
        const exported = createIdentifier(node.name);
        return {
            type: "ExportSpecifier",
            exported: exported,
            local: node.propertyName ? createIdentifier(node.propertyName) : exported,
        };
    }
}

let SOURCEMAPPING_URL = "sourceMa";
SOURCEMAPPING_URL += "ppingURL";
const SOURCEMAPPING_URL_RE = new RegExp(`^//#\\s+${SOURCEMAPPING_URL}=.+\\n?`, "m");
const COMPILERCACHE = new Map();
function createCompiler(options) {
    const file = path.basename(options.tsconfig);
    const configFileName = findConfigFile(options.tsconfig, sys.fileExists, path.extname(options.tsconfig) ? file : undefined);
    // istanbul ignore if
    if (!configFileName) {
        throw new Error(`rollup-plugin-dts: Couldn't find tsconfig file`);
    }
    const compilerOptions = Object.assign({}, options.compilerOptions, { noEmitOnError: false, noEmit: false, skipLibCheck: true, declaration: true, 
        // hm, there are still issues with this, I couldn’t get it to work locally
        // See https://github.com/Microsoft/TypeScript/issues/25662
        declarationMap: false });
    let diagnostic;
    const configParseResult = getParsedCommandLineOfConfigFile(configFileName, compilerOptions, Object.assign({}, sys, { onUnRecoverableConfigFileDiagnostic(d) {
            // istanbul ignore next
            diagnostic = d;
        } }));
    // istanbul ignore if
    if (!configParseResult) {
        console.log(diagnostic);
        throw new Error(`rollup-plugin-dts: Couldn't process compiler options`);
    }
    const { fileNames, options: parsedCompilerOptions } = configParseResult;
    return {
        parsedCompilerOptions,
        compiler: createProgram(fileNames, parsedCompilerOptions),
    };
}
function getCachedCompiler(options) {
    const cacheKey = JSON.stringify(options);
    function lazyCreate() {
        let compiler = COMPILERCACHE.get(cacheKey);
        if (!compiler) {
            compiler = createCompiler(options);
            COMPILERCACHE.set(cacheKey, compiler);
        }
        return compiler;
    }
    return {
        resolve(importee, importer) {
            const { parsedCompilerOptions } = lazyCreate();
            const result = nodeModuleNameResolver(importee, importer, parsedCompilerOptions, sys);
            // istanbul ignore else
            if (result.resolvedModule && result.resolvedModule.resolvedFileName) {
                return result.resolvedModule.resolvedFileName;
            }
            else {
                return null;
            }
        },
        transform(fileName) {
            const { compiler } = lazyCreate();
            const sourceFile = compiler.getSourceFile(fileName);
            let dtsFilename = "";
            let code = "";
            let map = `{"mappings": ""}`;
            const baseFileName = fileName.slice(0, -path.extname(fileName).length);
            compiler.emit(sourceFile, (fileName, data) => {
                if (fileName === `${baseFileName}.d.ts`) {
                    dtsFilename = fileName;
                    code = data.replace(SOURCEMAPPING_URL_RE, "").trim();
                }
                // if (fileName === `${baseFileName}.d.ts.map`) {
                //   map = data;
                // }
            });
            const dtsSource = createSourceFile(dtsFilename, code, ScriptTarget.Latest, true);
            const converter = new Transformer(dtsSource);
            const ast = converter.transform();
            // NOTE(swatinem):
            // hm, typescript generates `export default` without a declare,
            // but rollup moves the `export default` to a different place, which leaves
            // the function declaration without a `declare`.
            // Well luckily both words have the same length, haha :-D
            code = code.replace(/(export\s+)default(\s+(function|class))/m, "$1declare$2");
            return { code, ast, map };
        },
    };
}

// @ts-ignore
function dts(options = {}) {
    const filter = createFilter(options.include || ["*.ts+(|x)", "**/*.ts+(|x)"], options.exclude || []);
    const compiler = getCachedCompiler({
        tsconfig: options.tsconfig || process.cwd(),
        compilerOptions: options.compilerOptions || {},
    });
    return {
        name: "dts",
        resolveId(importee, importer) {
            if (!importer) {
                return null;
            }
            importer = importer.split("\\").join("/");
            return compiler.resolve(importee, importer);
        },
        async transform(_code, id) {
            // istanbul ignore if
            if (!filter(id)) {
                return;
            }
            return compiler.transform(id);
        },
    };
}

export default dts;
